/*
    Copyright (C) 2001-2003 KSVG Team
    This file is part of the KDE project

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#ifndef SVGAnimationElementImpl_H
#define SVGAnimationElementImpl_H

#include <qtimer.h>
#include <qptrlist.h>

#include "SVGElementImpl.h"
#include "SVGTestsImpl.h"
#include "SVGExternalResourcesRequiredImpl.h"

#include "ksvg_lookup.h"

namespace KSVG
{

class SVGTimer
{
friend class SVGAnimationElementImpl;
public:
	SVGTimer()
	{ timer = 0; ms = 0; invoked = false; singleShot = false; }
	~SVGTimer() { delete timer; }

	QTimer *timer;
	QPtrList<SVGAnimationElementImpl> notifyList;
	bool singleShot;
	bool invoked;
	int ms;
};

class SVGAnimationElementImpl : public SVGElementImpl,
								public SVGTestsImpl,
								public SVGExternalResourcesRequiredImpl
{
public:
	SVGAnimationElementImpl(DOM::ElementImpl *);
	virtual ~SVGAnimationElementImpl();

	SVGElementImpl *targetElement() const;
	void setTargetElement(SVGElementImpl *target);

	QTimer *addTimer(int ms, bool singleShot = true);
	void removeTimer(QTimer *timer);

	void applyAttribute(QString name, QString value, bool combine = false);

	virtual void handleTimerEvent(bool needCombine = false);
	virtual void setAttributes();

	int interval() const { return m_interval; }
	void setInterval(int interval) { m_interval = interval; }
	
	double getStartTime() const;
	double getCurrentTime() const;
	double getSimpleDuration() const;

	QString getRepeatCount() const { return m_repeatCount; }
	QString getRepeatDuration() const { return m_repeatDur; }

	QString getAttributeName() const { return m_attributeName; }
	QString getFrom() const { return m_from; }
	QString getTo() const { return m_to; }

private:
	double parseClockValue(const QString &data) const;
	
	SVGElementImpl *m_targetElement;
	QPtrList<SVGTimer> m_timerList;

	QString m_href;
	QString m_additive, m_accumulate;
	QString m_attributeName, m_attributeType;
	QString m_from, m_to, m_by;

	double m_begin, m_end, m_duration;
	int m_interval;
	
	QString m_calcMode;
	QString m_restart, m_repeatCount, m_repeatDur, m_fill;
	
	SVGStringListImpl *m_values, *m_keyTimes, *m_keySplines;
	
public:
	KSVG_GET
	KSVG_PUT

	enum
	{
		// Properties
		TargetElement,
		Href, Additive, Accumulate, AttributeName, AttributeType,
		CalcMode, Values, KeyTimes, KeySplines, From, To, By,
		Begin, Dur, End, Min, Max, Restart, RepeatCount, RepeatDur, Fill,
		// Functions
		GetStartTime, GetCurrentTime, GetSimpleDuration
	};

	KJS::Value getValueProperty(KJS::ExecState *exec, int token) const;
	void putValueProperty(KJS::ExecState *exec, int token, const KJS::Value &value, int attr);
};

};

KSVG_DEFINE_PROTOTYPE(SVGAnimationElementImplProto)
KSVG_IMPLEMENT_PROTOFUNC(SVGAnimationElementImplProtoFunc, SVGAnimationElementImpl)

#endif

// vim:ts=4:noet
