/*
    Copyright (C) 2001-2003 KSVG Team
    This file is part of the KDE project

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#ifndef SVGMaskElementImpl_H
#define SVGMaskElementImpl_H

#include "SVGTestsImpl.h"
#include "SVGBBoxTarget.h"
#include "SVGElementImpl.h"
#include "SVGStylableImpl.h"
#include "SVGLangSpaceImpl.h"
#include "SVGExternalResourcesRequiredImpl.h"

#include "ksvg_lookup.h"

namespace KSVG
{


class SVGRectImpl;
class SVGAnimatedLengthImpl;
class SVGAnimatedEnumerationImpl;
class SVGMaskElementImpl : public SVGElementImpl,
						   public SVGTestsImpl,
						   public SVGLangSpaceImpl,
						   public SVGExternalResourcesRequiredImpl,
						   public SVGStylableImpl,
						   public SVGBBoxTarget
{
public:
	SVGMaskElementImpl(DOM::ElementImpl *);
	virtual ~SVGMaskElementImpl();

	SVGAnimatedEnumerationImpl *maskUnits() const;
	SVGAnimatedEnumerationImpl *maskContentUnits() const;
	SVGAnimatedLengthImpl *x() const;
	SVGAnimatedLengthImpl *y() const;
	SVGAnimatedLengthImpl *width() const;
	SVGAnimatedLengthImpl *height() const;

	virtual SVGRectImpl *getBBox();
	virtual void setAttributes();
	
	virtual bool isContainer() const { return true; }
	virtual bool directRender() { return false; }

	void maskUnitsChanged();

private:
	SVGAnimatedEnumerationImpl *m_maskUnits;
	SVGAnimatedEnumerationImpl *m_maskContentUnits;
	SVGAnimatedLengthImpl *m_x;
	SVGAnimatedLengthImpl *m_y;
	SVGAnimatedLengthImpl *m_width;
	SVGAnimatedLengthImpl *m_height;

public:
	KSVG_GET
	KSVG_PUT
	KSVG_BRIDGE
	
	enum
	{
		// Properties
		MaskUnits, MaskContentUnits, X, Y, Width, Height
	};
	
	KJS::Value getValueProperty(KJS::ExecState *exec, int token) const;
	void putValueProperty(KJS::ExecState *exec, int token, const KJS::Value &value, int attr);
};

KSVG_REGISTER_ELEMENT(SVGMaskElementImpl, "mask")

};

#endif

// vim:ts=4:noet
